// Fragment Shader  file "basic_phong.frag"

#version 330

#extension GL_EXT_gpu_shader4 : enable

//Blinn-Phong Shading per Pixel for Multiple Light Sources

#define MAX_LIGHTS		3
#define MAX_FOG			0.6

//**********************************************************//
// Uniforms for shadow mapping
//**********************************************************//
uniform sampler2DArray shadowMap;	//array of shadow maps, because we have multiple lights
uniform int receive_shadow;			//skip shadowing to avoid self shadowing artifacts

//**********************************************************//
// Uniforms for texturing
//**********************************************************//
uniform sampler2D texture0;			//texture for this surface
uniform int texCount;				//number of textures (use other material if number is 0)

//**********************************************************//
// Uniforms for light sources
//**********************************************************//
uniform int light_counter;								//current number of lights (light loop)

uniform vec3 reflection_coefficient[MAX_LIGHTS];		//1. diffuse, 2. specular, 3.shininess
uniform vec3 attenuation[MAX_LIGHTS];					//attenuation
uniform vec3 light_color[MAX_LIGHTS];					//light color
uniform vec3 ambient_light[MAX_LIGHTS];					//ambient light
uniform vec4 light_position[MAX_LIGHTS];				//position of the light
uniform vec4 light_dir[MAX_LIGHTS];						//direction to the light
uniform mat4 lightViewMatrix[MAX_LIGHTS];				//transformation from world space to light space

uniform int spot_receive;								//to avoid artifacts not all surface receive lighting from spot lights

//**********************************************************//
// Uniforms for spotlights only
//**********************************************************//
uniform float spot_cos_cutoff[MAX_LIGHTS];				//cosinus cutoff 
uniform float spot_cone_height[MAX_LIGHTS];				//height of the spotlight cone
uniform float spot_scattering_effect[MAX_LIGHTS];		//scattering effect for the spot light cone 
uniform vec3  cameraPosition;							//position ot the camera (third column of the inverse model view matrix)

//**********************************************************//
// Uniforms for fog effect
//**********************************************************//
uniform vec4  fog_color;
uniform float fog_density;

//**********************************************************//
// Inputs from the vertex shader
//**********************************************************//
in vec4 vertex_position;				//position of the observer
in vec3 light_vector[MAX_LIGHTS];		//direction fragment to light
in vec3 half_vector[MAX_LIGHTS];		//half vector for lighting
in vec3 frag_color;						//surface color
in vec3 frag_normal;					//normal vector of fragment
in vec3 view_dir;						//direction fragment to camera
in vec2 tex_coord0;						//texture coordinates
in vec4 shadow_coord[MAX_LIGHTS];		//shadow coordinates for texture lookup

//**********************************************************//
// Final output color of fragment
//**********************************************************//
out vec4 fragColor;

//**********************************************************//
// Function headers
//**********************************************************//
float calculateShadow(int i, float weight);

float calculateSpotEffect(vec3 input_light_vector, float input_spot_cos_cutoff, vec3 input_light_dir);

float calculateScattering(vec3 input_start, vec3 input_dir, vec3 input_light_pos, float d);
void  calculateQuadraticEq(float a, float b, float c, out float minT, out float maxT);
void  calculateCone(vec3 rayOrigin, vec3 rayDir, mat4 invConeTransform, float aperture, float height, out float minT, out float maxT);

vec4  generateFog(vec4 fragColor);

//**********************************************************//
// Main program
//**********************************************************//
void main(void) {
	float shadow = 1.0;

	vec3 ambient_global = vec3(0.0, 0.0, 0.0);
	vec3 diffuse_global = vec3(0.0, 0.0, 0.0);
	vec3 specular_global = vec3(0.0, 0.0, 0.0);
	vec3 scatter_global = vec3(0.0, 0.0, 0.0);

	vec3 diffuse = vec3(0.0, 0.0, 0.0);
	vec3 specular = vec3(0.0, 0.0, 0.0);
	vec3 scatter = vec3(0.0, 0.0, 0.0);

	float spotEffect;

	for(int i=0; i<light_counter; i++)	{
		diffuse = vec3(0.0, 0.0, 0.0);
		specular = vec3(0.0, 0.0, 0.0);
		scatter = vec3(0.0, 0.0, 0.0);

		//**********************************************************//
		// Spot lights with light scattering effect
		//**********************************************************//
		if(light_dir[i].w == 1)	{
			shadow = calculateShadow(i, 0.05f);

			if(spot_receive == 1)	{
				//**********************************************************//
				// Light scattering
				//**********************************************************//
				float minT = 0.0;
				float maxT = 0.0;

				calculateCone(cameraPosition, normalize(view_dir), lightViewMatrix[i], spot_cos_cutoff[i], spot_cone_height[i], minT, maxT);

				maxT = clamp(maxT, 0.0, length(view_dir));
				minT = max(0.0, minT);

				float t = max(0.0, maxT - minT);

				float innerMin = (minT + maxT) * 0.5 - t * 0.6 * 0.5;

				if(spot_scattering_effect[i] > 0.0f)	{
					scatter = light_color[i] * calculateScattering(cameraPosition + normalize(view_dir) * minT, normalize(view_dir), light_position[i].xyz, t) * spot_scattering_effect[i];
				}

				//**********************************************************//
				// Spot effect
				//**********************************************************//
				spotEffect = calculateSpotEffect(normalize(light_vector[i]), spot_cos_cutoff[i], -light_dir[i].xyz);

				//**********************************************************//
				// Diffuse part (N_L = lambert term) (Spot light)
				//**********************************************************//
				float N_L = clamp(dot(frag_normal, normalize(light_vector[i])), 0.0, 1.0);

				//TODO: Blending
				//if(N_L > 0) {
					//**********************************************************//
					// Attenuation
					//**********************************************************//
					float distanceToLight = length(light_vector[i]);
	
					//constantAttenuation + linearAttenuation * distance + quadraticAttenuation * distance * distance
					vec3 attenuation = light_color[i] * 
						(1/(attenuation[i].x + attenuation[i].y * distanceToLight + attenuation[i].z * distanceToLight * distanceToLight));

					diffuse = reflection_coefficient[i].x * attenuation * N_L/(distanceToLight * distanceToLight);

					//**********************************************************//
					// Specular part (Spot light)
					//**********************************************************//
					float N_H = dot(half_vector[i], frag_normal);

					if(N_H > 0) {
						N_H = clamp(N_H, 0.0, 1.0);

						float highlight = pow(N_H, reflection_coefficient[i].z);

						specular = reflection_coefficient[i].y * (attenuation * highlight);
					}
				//}

				//add shadow only when the object is in the spot
				if(spotEffect <= 0)	{
					shadow = 1.0f;
				}

				ambient_global = ambient_global + ambient_light[i] * shadow * spotEffect;	//ambient part (in this implementation only for the spot)
				diffuse_global = diffuse_global + diffuse * shadow * spotEffect;			//diffuse part depends on spot effect
				specular_global = specular_global + specular * shadow * spotEffect;			//specular part depends on spot effect
				scatter_global = scatter_global + scatter;
			}
		}
		//**********************************************************//
		// Directional lights (no attenuation, etc.)
		//**********************************************************//
		else	{
			shadow = calculateShadow(i, 0.5f);

			//**********************************************************//
			// Diffuse part (N_L = lambert term) (Directional light)
			//**********************************************************//
			float N_L = max(dot(normalize(frag_normal), normalize(light_vector[i])), 0.0);

			if(N_L > 0) {
				diffuse = reflection_coefficient[i].x * (light_color[i] * N_L);

				diffuse = clamp(diffuse, 0.0, 1.0); 

				//**********************************************************//
				// Specular part (Directional light)
				//**********************************************************//
				float N_H = dot(frag_normal, normalize(normalize(view_dir) + normalize(light_vector[i])));

				if(N_H > 0) {
					float highlight = pow(N_H, reflection_coefficient[i].z);
			
					specular = reflection_coefficient[i].y * (light_color[i] * highlight);
				}
			}

			ambient_global = ambient_global + ambient_light[i] * shadow;		//ambient part
			diffuse_global = diffuse_global + diffuse * shadow;					//diffuse part
			specular_global = specular_global + specular * shadow;				//specular part
		}
	}

	//**********************************************************//
	// Apply blinn phong illumination model
	//**********************************************************//
	if(texCount > 0)	{
		//texture part
		vec3 textureColor = texture(texture0, tex_coord0.st).rgb;

		fragColor = vec4((ambient_global + diffuse_global + specular_global + scatter_global) * textureColor, 1.0);
	}
	else	{
		//only material
		fragColor = vec4((ambient_global + diffuse_global + specular_global + scatter_global) * frag_color, 1.0);
	}

	//**********************************************************//
	// Fog generation
	//**********************************************************//
	fragColor = generateFog(fragColor);
}

//**********************************************************//
// Fog generation
//**********************************************************//
vec4 generateFog(vec4 fragColor)	{
	const float LOG2 = 1.442695;

	float z = length(vertex_position);
	
	float fogFactor = exp2(-fog_density * fog_density * z * z * LOG2);
	
	fogFactor = clamp(fogFactor, 0.0, 1.0);

	fogFactor = max(fogFactor, MAX_FOG);

	return mix(fog_color, fragColor, fogFactor);
}

//**********************************************************//
// Calculate shadow
//**********************************************************//
float calculateShadow(int i, float weight)	{
	float shadow = 1.0;

	if(receive_shadow > 0 && shadow_coord[i].w > 0.0)	{
		vec4 shadowCoordinate = shadow_coord[i]/shadow_coord[i].w;
	
		shadowCoordinate.z += 0.0001;
	
		float distanceFromLight = texture2DArray(shadowMap, vec3(shadowCoordinate.st, i)).z;

		shadow = distanceFromLight < shadowCoordinate.z ? weight : 1.0;
	}

	return shadow;
}

//**********************************************************//
// Calculate spot effect
//**********************************************************//
float calculateSpotEffect(vec3 input_light_vector, float input_spot_cos_cutoff, vec3 input_light_dir)	{
	float a = 1.0 - acos(clamp(dot(input_light_vector, input_light_dir), 0.0, 1.0)) / input_spot_cos_cutoff;

	return clamp(a, 0.0, 1.0);
}

//**********************************************************//
// Calculate light scattering
//**********************************************************//
float calculateScattering(vec3 input_start, vec3 input_dir, vec3 input_light_pos, float d)	{
	vec3 q = input_start - input_light_pos;

	float b = dot(input_dir, q);
	float c = dot(q, q);

	float s = 1.0f / sqrt(c - b * b);

	float l = s * (atan((d + b) * s) - atan(b * s));

	return l;	
}

//**********************************************************//
// Solve quadratic equation
//**********************************************************//
void calculateQuadraticEq(float a, float b, float c, out float minT, out float maxT)	{
	float discriminant = b * b - 4.0 * a * c;

	if(discriminant < 0.0)	{
		maxT = 0.0;
		minT = 0.0;

		return;
	}

	float t = -0.5 * (b + sign(b) * sqrt(discriminant));
	float closestT = t / a;
	float furthestT = c / t;

	if(closestT > furthestT)	{	
		minT = furthestT;
		maxT = closestT;		
	}
	else	{
		minT = closestT;
		maxT = furthestT;
	}
}

//**********************************************************//
// Calculate a cone for light scattering
//**********************************************************//
void calculateCone(vec3 rayOrigin, vec3 rayDir, mat4 invConeTransform, float aperture, float height, out float minT, out float maxT)	{
	vec4 localOrigin = invConeTransform * vec4(rayOrigin, 1.0);
	vec4 localDir = invConeTransform * vec4(rayDir, 0.0);

	float tanTheta = tan(aperture);
	tanTheta *= tanTheta;

	float a = localDir.x * localDir.x + localDir.z * localDir.z - localDir.y * localDir.y * tanTheta;
	float b = 2.0 * (localOrigin.x * localDir.x + localOrigin.z * localDir.z - localOrigin.y * localDir.y * tanTheta);
	float c = localOrigin.x * localOrigin.x + localOrigin.z * localOrigin.z - localOrigin.y * localOrigin.y * tanTheta;

	calculateQuadraticEq(a, b, c, minT, maxT);

	float y1 = localOrigin.y + localDir.y * minT;
	float y2 = localOrigin.y + localDir.y * maxT;

	if(y1 > 0.0 && y2 > 0.0)	{
		minT = 0.0;
		maxT = -1.0;
	}
	else if(y1 > 0.0 && y2 < 0.0)	{
		minT = maxT;
		maxT = 10000.0;
	}
	else if(y1 < 0.0 && y2 > 0.0)	{
		maxT = minT;
		minT = 0.0;		
	}	
}